from pycont import Template, Contract
import unittest
import trafaret as t


class TestGenerators(unittest.TestCase):
    def test_simple_value(self):
        template = Template(t.Int())
        contract = Contract(template)
        result = contract.__call__(42)
        self.assertEqual(result, 42)

        with self.assertRaises(ValueError):
            result = contract('error')

    def test_simple_value_with_default(self):
        template = Template(t.Int(), default=7)
        contract = Contract(template)
        result = contract(42)
        self.assertEqual(result, 42)

        result = contract('error')
        self.assertEqual(result, 7)

    def test_list_single_value(self):
        template = Template(t.Int())
        contract = Contract([template])
        result = contract([42])
        self.assertEqual(result, [42])

        result = contract([8, 800, 555, 35, 35])
        self.assertEqual(result, [8, 800, 555, 35, 35])

        with self.assertRaises(ValueError):
            result = contract([8, 800, 555, 'error', 35])
        with self.assertRaises(ValueError):
            result = contract(42)

    def test_list_single_value_with_default(self):
        template = Template(t.Int(), 42)
        contract = Contract([template])

        result = contract([8, 800, 555, 35, 35])
        self.assertEqual(result, [8, 800, 555, 35, 35])

        result = contract([8, 800, 555, 'error', 35])
        self.assertEqual(result, [8, 800, 555, 42, 35])

    def test_list_many_values(self):
        contract = Contract([
            Template(t.Int()),
            Template(t.String()),
            Template(t.Float())
        ])
        result = contract([1, 'Test', 12.5])
        self.assertEqual(result, [1, 'Test', 12.5])

        with self.assertRaises(ValueError):
            result = contract(['error', 'Test', 12.5])
        with self.assertRaises(ValueError):
            result = contract([1, 666, 12.5])
        with self.assertRaises(ValueError):
            result = contract([1, 'Test', 'error'])
        with self.assertRaises(ValueError):
            result = contract([1, 'Test'])
        with self.assertRaises(ValueError):
            result = contract([1, 'Test', 12.5, 'error'])
        with self.assertRaises(ValueError):
            result = contract(42)

    def test_list_many_values_with_default(self):
        contract = Contract([
            Template(t.Int(), 42),
            Template(t.String(), 'null'),
            Template(t.Float(), 1.5)
        ])
        result = contract([1, 'Test', 12.5])
        self.assertEqual(result, [1, 'Test', 12.5])

        result = contract(['error', 'Test', 12.5])
        self.assertEqual(result, [42, 'Test', 12.5])
        result = contract([1, 666, 12.5])
        self.assertEqual(result, [1, 'null', 12.5])
        result = contract([1, 'Test', 'error'])
        self.assertEqual(result, [1, 'Test', 1.5])
        with self.assertRaises(ValueError):
            result = contract([1, 'Test'])
        with self.assertRaises(ValueError):
            result = contract([1, 'Test', 12.5, 'error'])
        with self.assertRaises(ValueError):
            result = contract(42)

    def test_dict_value(self):
        contract = Contract({
            'id': Template(t.Int()),
            'value': Template(t.String()),
        })
        result = contract({'id': 42, 'value': 'test'})
        self.assertEqual(result, {'id': 42, 'value': 'test'})
        result = contract({'id': 42, 'value': 'test', 'key': 666})
        self.assertEqual(result, {'id': 42, 'value': 'test'})

        with self.assertRaises(ValueError):
            result = contract({'id': 'error', 'value': 'test'})
        with self.assertRaises(ValueError):
            result = contract({'id': 42, 'value': 666})
        with self.assertRaises(ValueError):
            result = contract({'id': 1})
        with self.assertRaises(ValueError):
            result = contract(666)

    def test_dict_value_with_default(self):
        contract = Contract({
            'id': Template(t.Int(), 99),
            'value': Template(t.String(), 'null'),
        })
        result = contract({'id': 42, 'value': 'test'})
        self.assertEqual(result, {'id': 42, 'value': 'test'})
        result = contract({'id': 42, 'value': 'test', 'key': 666})
        self.assertEqual(result, {'id': 42, 'value': 'test'})

        result = contract({'id': 'error', 'value': 'test'})
        self.assertEqual(result, {'id': 99, 'value': 'test'})
        result = contract({'id': 42, 'value': 666})
        self.assertEqual(result, {'id': 42, 'value': 'null'})
        result = contract({'id': 99})
        self.assertEqual(result, {'id': 99, 'value': 'null'})
        with self.assertRaises(ValueError):
            result = contract(666)

    def test_binary_template(self):
        # Simple values
        template_1 = Template(t.Int())
        template_2 = Template(t.String())
        contract = Contract(template_1 | template_2)
        result = contract(42)
        self.assertEqual(result, 42)
        result = contract('test')
        self.assertEqual(result, 'test')
        with self.assertRaises(ValueError):
            result = contract(['list'])

        # Simple list values
        template_1 = Template(t.Int())
        template_2 = Template(t.String())
        contract = Contract([template_1 | template_2])
        result = contract([1, 2, 3])
        self.assertEqual(result, [1, 2, 3])
        result = contract(['test_1', 'test_2'])
        self.assertEqual(result, ['test_1', 'test_2'])
        result = contract([1, 'test', 3])
        self.assertEqual(result, [1, 'test', 3])

        # Combinated list values
        template_1 = Template(t.Int())
        template_2 = Template(t.String())
        contract = Contract([
            template_1 | template_2,
            template_1,
            template_2
        ])
        result = contract([1, 1, 'test'])
        self.assertEqual(result, [1, 1, 'test'])
        result = contract(['test', 1, 'test'])
        self.assertEqual(result, ['test', 1, 'test'])

        # Dict values
        template = {
            'key': Template(t.String()) | Template(t.Int()),
            'value': Template(t.String()),
        }
        contract = Contract(template)
        result = contract({'key': 'key', 'value': 'test'})
        self.assertEqual(result, {'key': 'key', 'value': 'test'})
        result = contract({'key': 12, 'value': 'test'})
        self.assertEqual(result, {'key': 12, 'value': 'test'})

        template_1 = {
            'key': Template(t.String()),
            'value': Template(t.String()),
        }
        template_2 = {
            'id': Template(t.Int()),
            'value': Template(t.String()),
        }
        contract_1 = Contract(template_1)
        contract_2 = Contract(template_2)
        contract = contract_1 | contract_2
        result = contract({'key': 'key', 'value': 'test'})
        self.assertEqual(result, {'key': 'key', 'value': 'test'})
        result = contract({'id': 42, 'value': 'test'})
        self.assertEqual(result, {'id': 42, 'value': 'test'})

    def test_optional(self):
        contract = Contract({
            'key': Template(t.Int()),
            'value': Template(t.String()),
            'optional': Template(t.String(), optional=True)
        })
        data = {
            'key': 1,
            'value': 'test',
            'optional': 'test',
        }
        result = contract(data)
        self.assertEqual(result, data)
        del data['optional']
        result = contract(data)
        self.assertEqual(result, data)

    def test_complex_dict_value(self):
        contract = Contract({
            'key': Template(t.Int()),
            'value': Template(t.String()),
            'optional': Template(t.String(), optional=True),
            'property': [Template(t.String())],
            'optional_property': [Template(t.String(), optional=True)],
            'objects': [{'id': Template(t.Int()), 'val': Template(t.String())}],
        }, optional_keys=['optional_property'])
        data = {
            'key': 1,
            'value': 'test',
            'optional': 'test',
            'property': ['A', 'B', 'C'],
            'optional_property': ['A', 'B', 'C'],
            'objects': [{'id': 1, 'val': 'test'}, {'id': 2, 'val': 'test'}],
        }
        result = contract(data)
        self.assertEqual(result, data)
        del data['optional_property']
        del data['optional']
        result = contract(data)
        self.assertEqual(result, data)

        del data['objects']
        with self.assertRaises(ValueError):
            result = contract(data)



from pycont import Contract, Template, TemplateError

import unittest
import trafaret as t


class TestValidator(unittest.TestCase):
    def test_template(self):
        with self.assertRaises(ValueError):
            template = Template('error')

        trafaret = t.String()
        template = Template(trafaret)
        self.assertEqual(template.template, [trafaret])
        template.check('test')
        with self.assertRaises(ValueError):
            template.check(42)

        template.template = t.Int()
        template.check(42)

        del template.template
        with self.assertRaises(ValueError):
            template.check('test')

    def test_default(self):
        trafaret = t.String()
        template = Template(trafaret, 'null')
        self.assertEqual(template.default, 'null')
        template.default = 'new'
        self.assertEqual(template.default, 'new')

        with self.assertRaises(ValueError):
            template.default = 42

        del template.template
        with self.assertRaises(ValueError):
            template.default = 42

        del template.default
        self.assertIsNone(template.default)

    def test_simple_validator(self):
        # Create contract
        trafaret = t.Trafaret()
        template = Template(trafaret)
        contract = Contract(template)
        self.assertEqual(contract.template, template)
        new_template = Template(trafaret)
        contract.template = new_template
        self.assertEqual(contract.template, new_template)

        # Int
        int_t = Template(t.Int())
        contract = Contract(int_t)

        # String
        string_t = Template(t.String())
        contract = Contract(string_t)

        # Dict
        dict_t = Template(t.Dict({
            'id': t.Int,
            'email': t.Email
        }))
        contract.template = dict_t

        # List
        list_t = Template(t.List(t.Int))
        contract.template = list_t

    def test_list_validator(self):
        list_t = [
            Template(t.Int()),
            Template(t.List(t.Int)),
            Template(t.Dict({'id': t.Int, 'val': t.String})),
        ]
        contract = Contract(list_t)

        list_t = [
            Template(t.Int()),
            'error',
            Template(t.Dict({'id': t.Int, 'val': t.String})),
        ]
        with self.assertRaises(ValueError):
            contract.template = list_t
        with self.assertRaises(ValueError):
            contract = Contract(list_t)

    def test_dict_validator(self):
        dict_t = {
            'id': Template(t.Int(gt=0)),
            'val': Template(t.String()),
        }
        contract = Contract(dict_t)

        dict_t = {
            'id': Template(t.Int(gt=0)),
            'val': 'error',
        }
        with self.assertRaises(ValueError):
            contract.template = dict_t
        with self.assertRaises(ValueError):
            contract = Contract(dict_t)

        dict_t = {
            12: Template(t.Int(gt=0)),
            'val': 'error',
        }
        with self.assertRaises(TypeError):
            contract = Contract(dict_t)

    def test_binary_templates(self):
        tempalte_1 = Template(t.Int())
        tempalte_2 = Template(t.String())
        contract = Contract(tempalte_1 | tempalte_2)
        result = contract(42)
        self.assertEqual(result, 42)
        result = contract('test')
        self.assertEqual(result, 'test')
        with self.assertRaises(ValueError):
            result = contract([123])

        tempalte_1 = Template(t.Int(), default=42)
        tempalte_2 = Template(t.String(), default='Test')
        with self.assertRaises(TemplateError):
            tempalte = tempalte_1 | tempalte_2

        tempalte_1 = Template(t.Int(), default=42)
        tempalte_2 = Template(t.String())
        tempalte = tempalte_1 | tempalte_2  # noqa

        tempalte_1 = Template(t.Int())
        tempalte_2 = Template(t.String(), default='Test')
        tempalte = tempalte_1 | tempalte_2  # noqa



from typing import Any, Union, List, Dict, TypeVar, Optional
from trafaret import DataError
from pycont.template import Template


T = TypeVar('Contract')


class Contract:
    """
    Contract class contains template to validate and generate object if data is valid

    Args:
    - template Union[Template, List[Template], Dict[str, Template]]:
        Template object or list of Templates or dict of Templates
    - optional_keys Optional[List[str]] = None
        Used only in dictionary contract
        The check function does not raise an error if the key is not set.

    Raises:
    - ValueError if template is not valid

    Simple value validation and generate
        >>> from pycont import Contract, Template
        >>> import trafaret as t
        >>> contract = Contract(Template(t.Int()))
        >>> print(contract(42))
        42
        >>> print(contract('test'))
        ValueError: invalid literal for int() with base 10: 'test'
    """
    def __init__(
        self,
        template: Union[Template, List[Template], Dict[str, Template]],
        optional_keys: Optional[List[str]] = None,
    ):
        self._validate(template)
        self._templates = [template]
        self.optional_keys = optional_keys or []

    def __or__(self, other: T) -> T:
        self._templates = self._templates + other._templates
        return self

    template = property()

    @template.getter
    def template(self) -> Union[Template, List[Template], Dict[str, Template]]:
        """
        Get current template value
        Return:
            Current template or list of tempalates
        """
        return self._templates[0] if len(self._templates) == 1 else self._templates[0]

    @template.setter
    def template(
        self,
        template: Union[Template, List[Template], Dict[str, Template]]
    ) -> None:
        """
        Set new template value

        Args:
            template Union[Template, List[Template], Dict[str, Template]]:
                new Template object or list of Template or Dict[str, Template]
        Raises:
            ValueError if template is not valid
        """
        self._validate(template)
        self._templates = [template]

    def _validate(
        self,
        template: Union[Template, List[Template], Dict[str, Template]]
    ):
        """
        Check what seted template is valid

        Raises:
            ValueError if template is not valid
        """
        if isinstance(template, list):
            for value in template:
                self._validate(value)
            return
        if isinstance(template, dict):
            for key, value in template.items():
                if not isinstance(key, str):
                    raise TypeError(f'Invalid key "{key}" type: key in template must be string')
                self._validate(value)
            return
        if not isinstance(template, Template):
            raise ValueError('Invalid value: value type must be Template')

    def _check(
        self,
        template: Union[Template, List[Template], Dict[str, Template]],
        data: Any
    ):
        """
        Recursive check if seted value is valid by template

        Args:
            template Union[Template, List[Template], Dict[str, Template]:
                Template object or list of Templates or dict of Templates
            data Any: data for checking

        Raises:
            trafaret.DataError if data cant checked
            ValueError or more Exepitons if cant read or validate data
        """
        if isinstance(template, list):
            result = []
            if len(template) == 1:
                sub_template = template[0]
                for value in data:
                    result.append(self._check(sub_template, value))
            else:
                if len(template) != len(data):
                    raise ValueError("Invalid value length")
                for inx, sub_template in enumerate(template):
                    next_data = data[inx]
                    result.append(self._check(sub_template, next_data))
            return result

        if isinstance(template, dict):
            result = {}
            for key, sub_template in template.items():
                if key in data.keys():
                    result[key] = self._check(sub_template, data[key])
                else:
                    if key not in self.optional_keys:
                        if isinstance(sub_template, Template):
                            if sub_template.default is not None:
                                result[key] = sub_template.default
                            else:
                                if not sub_template.optional:
                                    raise ValueError(f'Key "{key}" not set')
                        else:
                            raise ValueError(f'Key "{key}" not set')
            return result
        try:
            template.check(data)
        except DataError as e:
            if template.default is not None:
                data = template.default
            else:
                raise e
        return data

    def __call__(self, data: Any) -> Any:
        """
        Validate and generate object by template

        Args:
            data Any: data for validation

        Return:
            Any: data if data is valid

        Raises:
            ValueError if data is not valid
        """
        errors = []
        for template in self._templates:
            try:
                result = self._check(template, data)
                return result
            except Exception as e:
                errors.append(e)
        raise ValueError(f"Invalid value: {errors}")



from typing import Any, Union, List, Dict
from trafaret import DataError

from pycont import Contract
from pycont.asyncio.template import AsyncTemplate as Template


class AsyncContract(Contract):
    """
    Contract class contains template to validate and generate object if data is valid

    Args:
    - template Union[Template, List[Template], Dict[str, Template]]:
        Template object or list of Templates or dict of Templates

    Raises:
    - ValueError if template is not valid

    Simple value validation and generate
        >>> from pycont import AsyncContract as Contract
        >>> from pycont import AsyncTemplate as Tempalte
        >>> import trafaret as t
        >>> contract = Contract(Template(t.Int()))
        >>> await contract(42)
    """
    async def _check(
        self,
        template: Union[Template, List[Template], Dict[str, Template]],
        data: Any
    ):
        """
        Recursive check if seted value is valid by template

        Args:
            template Union[Template, List[Template], Dict[str, Template]:
                Template object or list of Templates or dict of Templates
            data Any: data for checking

        Raises:
            trafaret.DataError if data cant checked
            ValueError or more Exepitons if cant read or validate data
        """
        if isinstance(template, list):
            result = []
            if len(template) == 1:
                sub_template = template[0]
                for value in data:
                    result.append(await self._check(sub_template, value))
            else:
                if len(template) != len(data):
                    raise ValueError("Invalid value length")
                for inx, sub_template in enumerate(template):
                    next_data = data[inx]
                    result.append(await self._check(sub_template, next_data))
            return result

        if isinstance(template, dict):
            result = {}
            for key, sub_template in template.items():
                if key in data.keys():
                    result[key] = await self._check(sub_template, data[key])
                else:
                    if key not in self.optional_keys:
                        if isinstance(sub_template, Template):
                            if sub_template.default is not None:
                                result[key] = sub_template.default
                            else:
                                if not sub_template.optional:
                                    raise ValueError(f'Key "{key}" not set')
                        else:
                            raise ValueError(f'Key "{key}" not set')
            return result
        try:
            await template.check(data)
        except DataError as e:
            if template.default is not None:
                data = template.default
            else:
                raise e
        return data

    async def __call__(self, data: Any) -> Any:
        """
        Validate and generate object by template

        Args:
            data Any: data for validation

        Return:
            Any: data if data is valid

        Raises:
            ValueError if data is not valid
        """
        errors = []
        for template in self._templates:
            try:
                result = await self._check(template, data)
                return result
            except Exception as e:
                errors.append(e)
        raise ValueError(f"Invalid value: {errors}")



from trafaret import Trafaret, DataError
from typing import Any, Optional, Union, List, TypeVar


T = TypeVar('Template')


class TemplateError(Exception):
    pass


class Template():
    """
    Template object to validate on contract class
    May contain a default value returned if the data is invalid

    Args:
        template trafarer.Trafaret objects or subclass:
            The template for checking value
        default Optional[Any] = None:
            The default value returned if the data is invalid
            The default value type mast be valid for the template object
        optional Optional[bool] = False:
            Used only in dictionary contract
            If True, the check function does not raise an error if the key is not set.

        >>> from pycont import Contract, Template
        >>> import trafaret as t
        >>> contract = Contract(Template(t.Int()))
    """
    def __init__(
        self,
        template: Union[Trafaret, List[Trafaret]],
        default: Optional[Any] = None,
        optional: Optional[bool] = False,
    ):
        self._validate_template(template)
        self._template = [template] if isinstance(template, Trafaret) else template
        self.optional = optional

        if default is not None:
            self._validate_default(default)
        self._default = default

    def __or__(self, other: T):
        if self.default is not None and other.default is not None:
            raise TemplateError('Both templates cannot have default values')
        self.template = self.template + other.template
        self.default = self.default or other.default
        return self

    template = property()

    @template.getter
    def template(self) -> Trafaret:
        """
        Get current template

        Return:
            trafaret.Trafaret: current template
        """
        return self._template

    @template.setter
    def template(self, template: Union[Trafaret, List[Trafaret]]) -> None:
        """
        Validate new trafaret object
        Args:
            template: trafarer.Trafaret objects or subclass
        """
        self._validate_template(template)
        self._template = [template] if isinstance(template, Trafaret) else template

    @template.deleter
    def template(self):
        """
        Delete old template
        """
        self._template = None

    def _validate_template(self, template: Trafaret) -> None:
        """
        Check if template is valid for work
        """
        if not isinstance(template, Trafaret):
            try:
                for tmp in template:
                    if not isinstance(tmp, Trafaret):
                        raise ValueError(f'Template type must be trafaret.Trafaret, not {type(template)}')
            except Exception:
                raise ValueError(f'Template type must be trafaret.Trafaret, not {type(template)}')

    default = property()

    @default.getter
    def default(self) -> Trafaret:
        """
        Get current default value
        """
        return self._default

    @default.setter
    def default(self, default: Any) -> None:
        """
        Validate and set new default value

        Args:
            default Any:
                The default value returned if the data is invalid
                The default value type mast be valid for the template object
        """
        self._validate_default(default)
        self._default = default

    @default.deleter
    def default(self):
        """
        Delete default value
        """
        self._default = None

    def _validate_default(self, default):
        """
        Check if default value is valid
        Raises:
            ValueError if template is not set
            trafaret.DataError if value is not valid
        """
        if default is None:
            return
        if self._template is not None:
            for template in self._template:
                try:
                    template.check(default)
                except DataError:
                    pass
                else:
                    return
        raise ValueError("Template not set")

    def check(self, value: Any):
        """
        Check if value is valid by template

        Raises:
            ValueError if template is not set
            trafaret.DataError if value is not valid
        """
        if self._template is None:
            raise ValueError("Template not set")
        errors = []
        for template in self._template:
            try:
                template.check(value)
            except DataError as e:
                errors.append(e.error)
            else:
                return
        raise DataError(error='\n'.join(errors))



from .contract import Contract  # noqa
from .template import Template, TemplateError  # noqa
from .asyncio.contract import AsyncContract  # noqa
from .asyncio.template import AsyncTemplate  # noqa

__all__ = (
    "Contract",
    "Template",

    "TemplateError",

    "AsyncContract",
    "AsyncTemplate"
)

__version__ = '0.0.4'



from trafaret import DataError
from typing import Any

from pycont import Template


class AsyncTemplate(Template):
    """
    Template object to validate on contract class
    May contain a default value returned if the data is invalid

    Args:
        template trafarer.Trafaret objects or subclass:
            The template for checking value
        default Optional[Any] = None:
            The default value returned if the data is invalid
            The default value type mast be valid for the template object

        >>> from pycont import Contract, Template
        >>> import trafaret as t
        >>> contract = Contract(Template(t.Int()))
    """
    async def check(self, value: Any):
        """
        Check if value is valid by template

        Raises:
            ValueError if template is not set
            trafaret.DataError if value is not valid
        """
        if self._template is None:
            raise ValueError("Template not set")
        errors = []
        for template in self._template:
            try:
                await template.async_check(value)
            except DataError as e:
                errors.append(e.error)
            else:
                return
        raise DataError(error='\n'.join(errors))



from setuptools import setup, find_packages
import re

with open('requirements.txt') as f:
    required = f.read().splitlines()

pattern = re.compile(r'__version__\s=\s[\"\'](\d*.\d*.\d*)[\"\']')
with open('pycont/__init__.py') as f:
    version = pattern.findall(f.read())[0]

setupconf = dict(
    name='pycont',
    url='https://github.com/AlexeyPichugin/pycont',
    license='MIT License',
    version=version,
    author='Alexey Pichugin',
    author_email="a.o.pichugin@outlook.com",
    description='Validate and generate data from templates',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Text Processing :: Filters',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=find_packages(),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    package_data={
        '': ['requirements.txt']
    },
    install_requires=required,
    zip_safe=False,
)

if __name__ == '__main__':
    setup(**setupconf)
